@@ -1,6 +1,6 @@
 /*
  * Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
- * SPDX-License-Identifier: Apache-2.0.
+ * SPDX-License-Identifier: Apache-2.0
  */
 
 use bytes::Bytes;
@@ -13,6 +13,9 @@ use std::pin::Pin;
 use std::sync::Arc;
 use std::task::{Context, Poll};
 
+use crate::callback::BodyCallback;
+use crate::header::append_merge_header_maps;
+
 pub type Error = Box<dyn StdError + Send + Sync>;
 
 /// SdkBody type
@@ -32,6 +35,9 @@ pub struct SdkBody {
     /// In the event of retry, this function will be called to generate a new body. See
     /// [`try_clone()`](SdkBody::try_clone)
     rebuild: Option<Arc<dyn (Fn() -> Inner) + Send + Sync>>,
+    /// A list of callbacks that will be called at various points of this `SdkBody`'s lifecycle
+    #[pin]
+    callbacks: Vec<Box<dyn BodyCallback>>,
 }
 
 impl Debug for SdkBody {
@@ -74,6 +80,7 @@ impl SdkBody {
         Self {
             inner: Inner::Dyn(body),
             rebuild: None,
+            callbacks: Vec::new(),
         }
     }
 
@@ -90,6 +97,7 @@ impl SdkBody {
         SdkBody {
             inner: initial.inner,
             rebuild: Some(Arc::new(move || f().inner)),
+            callbacks: Vec::new(),
         }
     }
 
@@ -97,6 +105,7 @@ impl SdkBody {
         Self {
             inner: Inner::Taken,
             rebuild: None,
+            callbacks: Vec::new(),
         }
     }
 
@@ -104,6 +113,7 @@ impl SdkBody {
         Self {
             inner: Inner::Once(None),
             rebuild: Some(Arc::new(|| Inner::Once(None))),
+            callbacks: Vec::new(),
         }
     }
 
@@ -111,7 +121,8 @@ impl SdkBody {
         self: Pin<&mut Self>,
         cx: &mut Context<'_>,
     ) -> Poll<Option<Result<Bytes, Error>>> {
-        match self.project().inner.project() {
+        let mut this = self.project();
+        let polling_result = match this.inner.project() {
             InnerProj::Once(ref mut opt) => {
                 let data = opt.take();
                 match data {
@@ -125,7 +136,29 @@ impl SdkBody {
             InnerProj::Taken => {
                 Poll::Ready(Some(Err("A `Taken` body should never be polled".into())))
             }
+        };
+
+        match &polling_result {
+            // When we get some bytes back from polling, pass those bytes to each callback in turn
+            Poll::Ready(Some(Ok(bytes))) => {
+                for callback in this.callbacks.iter_mut() {
+                    // Callbacks can run into errors when reading bytes. They'll be surfaced here
+                    callback.update(bytes)?;
+                }
+            }
+            // When we're done polling for bytes, run each callback's `trailers()` method. If any calls to
+            // `trailers()` return an error, propagate that error up. Otherwise, continue.
+            Poll::Ready(None) => {
+                for callback_result in this.callbacks.iter().map(BodyCallback::trailers) {
+                    if let Err(e) = callback_result {
+                        return Poll::Ready(Some(Err(e)));
+                    }
+                }
+            }
+            _ => (),
         }
+
+        polling_result
     }
 
     /// If possible, return a reference to this body as `&[u8]`
@@ -143,9 +176,12 @@ impl SdkBody {
     pub fn try_clone(&self) -> Option<Self> {
         self.rebuild.as_ref().map(|rebuild| {
             let next = rebuild();
-            SdkBody {
+            let callbacks = self.callbacks.iter().map(BodyCallback::make_new).collect();
+
+            Self {
                 inner: next,
                 rebuild: self.rebuild.clone(),
+                callbacks,
             }
         })
     }
@@ -153,6 +189,11 @@ impl SdkBody {
     pub fn content_length(&self) -> Option<u64> {
         self.size_hint().exact()
     }
+
+    pub fn with_callback(&mut self, callback: Box<dyn BodyCallback>) -> &mut Self {
+        self.callbacks.push(callback);
+        self
+    }
 }
 
 impl From<&str> for SdkBody {
@@ -166,6 +207,7 @@ impl From<Bytes> for SdkBody {
         SdkBody {
             inner: Inner::Once(Some(bytes.clone())),
             rebuild: Some(Arc::new(move || Inner::Once(Some(bytes.clone())))),
+            callbacks: Vec::new(),
         }
     }
 }
@@ -175,6 +217,7 @@ impl From<hyper::Body> for SdkBody {
         SdkBody {
             inner: Inner::Streaming(body),
             rebuild: None,
+            callbacks: Vec::new(),
         }
     }
 }
@@ -212,7 +255,30 @@ impl http_body::Body for SdkBody {
         self: Pin<&mut Self>,
         _cx: &mut Context<'_>,
     ) -> Poll<Result<Option<HeaderMap<HeaderValue>>, Self::Error>> {
-        Poll::Ready(Ok(None))
+        let mut header_map = None;
+        // Iterate over all callbacks, checking each for any `HeaderMap`s
+        for callback in &self.callbacks {
+            match callback.trailers() {
+                // If this is the first `HeaderMap` we've encountered, save it
+                Ok(Some(right_header_map)) if header_map.is_none() => {
+                    header_map = Some(right_header_map);
+                }
+                // If this is **not** the first `HeaderMap` we've encountered, merge it
+                Ok(Some(right_header_map)) if header_map.is_some() => {
+                    header_map = Some(append_merge_header_maps(
+                        header_map.unwrap(),
+                        right_header_map,
+                    ));
+                }
+                // Early return if a callback encountered an error.
+                Err(e) => {
+                    return Poll::Ready(Err(e));
+                }
+                // Otherwise, continue on to the next iteration of the loop.
+                _ => continue,
+            }
+        }
+        Poll::Ready(Ok(header_map))
     }
 
     fn is_end_stream(&self) -> bool {
@@ -296,10 +362,9 @@ mod test {
         let _ = format!("{:?}", body);
     }
 
-    fn is_send<T: Send + Sync>() {}
-
     #[test]
     fn sdk_body_is_send() {
+        fn is_send<T: Send>() {}
         is_send::<SdkBody>()
     }
 }
